import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

//Builder interface to create a data when building the widget tree
typedef _ObserverInjectionBuilder<T> = T Function(BuildContext context);

//Dispose interface that is called when the ObserverInjection class is disposed from widget tree
typedef _Dispose<T> = void Function(BuildContext, T value);

//ObserverInjection class that hold an Object and change the widget three when this object changes
class ObserverInjection<T> extends StatefulWidget {
  final _ObserverInjectionBuilder<T> builder;
  final _Dispose<T> dispose;
  final Widget child;

  const ObserverInjection({
    Key key,
    @required this.builder,
    this.dispose,
    @required this.child,
  })  : assert(builder != null && child != null),
        super(key: key);

  static T of<T>(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType<_ObserverInjector<T>>().data;

  @override
  _ObserverInjectionState<T> createState() => _ObserverInjectionState<T>();
}

class _ObserverInjectionState<T> extends State<ObserverInjection<T>> {
  T data;

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    data = widget.builder(context);
  }

  @override
  void dispose() {
    if (widget.dispose != null) {
      widget.dispose(context, data);
    }
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _ObserverInjector<T>(
      child: widget.child,
      data: data,
    );
  }
}

//_ObserverInjector class that inject a value received from ObserverInjection class into the whole tree
class _ObserverInjector<T> extends InheritedWidget {
  const _ObserverInjector({
    Key key,
    @required this.data,
    @required Widget child,
  })  : assert(child != null),
        super(key: key, child: child);

  final T data;

  @override
  bool updateShouldNotify(_ObserverInjector<T> old) => old.data != data;
}
